@@ -1,24 +1,22 @@
 /* D-Bus Message unserialization. This contains all the logic to map from
  * D-Bus types to Python objects.
  *
- * Copyright (C) 2006 Collabora Ltd.
+ * Copyright (C) 2006, 2007 Collabora Ltd.
  *
  * Licensed under the Academic Free License version 2.1
  *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
+ * This library is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU Lesser General Public License as published by
+ * the Free Software Foundation; either version 2.1 of the License, or
  * (at your option) any later version.
  *
- * This program is distributed in the hope that it will be useful,
+ * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
+ * GNU Lesser General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
+ * You should have received a copy of the GNU Lesser General Public License
+ * along with this library; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  *
  */
@@ -120,9 +118,17 @@ _message_iter_get_dict(DBusMessageIter *iter,
     PyObject *ret;
     int status;
 
+    if (!sig_str) {
+        PyErr_NoMemory();
+        return NULL;
+    }
     sig = PyObject_CallFunction((PyObject *)&DBusPySignature_Type,
                                 "(s#)", sig_str+2,
                                 (Py_ssize_t)strlen(sig_str)-3);
+    dbus_free(sig_str);
+    if (!sig) {
+        return NULL;
+    }
     status = PyDict_SetItem(kwargs, dbus_py_signature_const, sig);
     Py_DECREF(sig);
     if (status < 0) {
@@ -236,7 +242,7 @@ _message_iter_get_pyobject(DBusMessageIter *iter,
                                     args, kwargs);
             }
             else {
-                args = Py_BuildValue("(O)", PyUnicode_DecodeUTF8(u.s,
+                args = Py_BuildValue("(N)", PyUnicode_DecodeUTF8(u.s,
                                                                  strlen(u.s),
                                                                  NULL));
                 if (!args) {
@@ -455,12 +461,8 @@ _message_iter_get_pyobject(DBusMessageIter *iter,
                          "message", type);
     }
 
-    if (args) {
-        Py_DECREF(args);
-    }
-    if (kwargs) {
-        Py_DECREF(kwargs);
-    }
+    Py_XDECREF(args);
+    Py_XDECREF(kwargs);
     return ret;
 }
 
