@@ -4,20 +4,18 @@
  *
  * Licensed under the Academic Free License version 2.1
  *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
+ * This library is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU Lesser General Public License as published by
+ * the Free Software Foundation; either version 2.1 of the License, or
  * (at your option) any later version.
  *
- * This program is distributed in the hope that it will be useful,
+ * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
+ * GNU Lesser General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
+ * You should have received a copy of the GNU Lesser General Public License
+ * along with this library; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  *
  */
@@ -113,13 +111,14 @@ _pending_call_notify_function(DBusPendingCall *pc,
                 PyErr_Print();
             }
             Py_XDECREF(ret);
+            Py_DECREF(msg_obj);
         }
         /* else OOM has happened - not a lot we can do about that,
          * except possibly making it fatal (FIXME?) */
     }
 
-    Py_XDECREF(handler);
 release:
+    Py_XDECREF(handler);
     PyGILState_Release(gil);
 }
 
@@ -163,7 +162,7 @@ DBusPyPendingCall_ConsumeDBusPendingCall(DBusPendingCall *pc,
     PyList_SET_ITEM(list, 0, callable);
 
     /* INCREF so we can give a ref to set_notify and still have one */
-    Py_INCREF(list);    
+    Py_INCREF(list);
 
     Py_BEGIN_ALLOW_THREADS
     ret = dbus_pending_call_set_notify(pc,
